
import os
import pathlib
import numpy as np
import windrose as wr
import random
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

from multiprocessing import Pool
from metpy.calc import wind_speed, wind_direction
from datetime import datetime
from typing import Literal, Optional
from collections.abc import Iterable

from ..queryreport import print_text
from ..utilities import nice_round_up
from .met import NetcdfMet

iVector = Iterable[list[int]]
fVector = Iterable[list[float]]
sVector = Iterable[list[str]]

seasonDict = dict[str,iVector]
seasonList = list[seasonDict]

_ENSEMBLE_TYPES = Literal["randomize", "sequential"]

class MetRepo:
    """
    A class to access a repository of netcdf met data.
    The repository is assumed to be a directory containing netcdf files in the form
    YYYYMM.nc
    with each file containing met data on pressure levels over days, hours, latitude and longitude

    """
    _lat = float
    _lon = float

    _dirc = str
    _filetype = str
    _files = sVector
    _years = iVector
    _months = iVector
    _days = iVector
    _hours = iVector

    _seasons = seasonList
    _seasons_combine = str

    def __init__(self, dirc: str, lat: float, lon: float, filetype: Literal["nc","grib"]="nc"):
        """
        Initialize a MetRepo instance

        :param dirc: path to met data repository
        :type dirc: str
        :param lat: latitude 
        :type lat: float
        :param lon: _description_
        :type lon: float
        :param filetype: _description_, defaults to "nc"
        :type filetype: Literal[&quot;nc&quot;,&quot;grib&quot;], optional
        :raises RuntimeError: if dirc is not a directory or is not found
        """

        if not os.path.isdir(dirc):
            raise RuntimeError('{dirc} is not a directory or not found')
        
        self._lat = lat
        self._lon = lon

        self._dirc = dirc
        self._filetype = filetype

        self._files = [file.stem for file in sorted(pathlib.Path(self.dirc).glob(f"*.{filetype}"))]

        self._years = []
        self._months = []
        self._days = []
        self._hours = []

        self._seasons = []
        self._seasons_combine = 'first'

    @property
    def latitude(self):
        return self._lat
    
    @property
    def longitude(self):
        return self._lon
    
    @property
    def years_available(self) -> set[int]:
        return set(sorted([int(f[:4]) for f in self._files]))
    
    @property
    def months_available(self) -> set[int]:
        return set(sorted([int(f[4:6]) for f in self._files]))

    @property
    def dirc(self):
        return self._dirc
    
    @property
    def years(self):
        return self._years

    @years.setter
    def years(self, yr: list[int]):
        self._years = yr

    @property
    def months(self):
        return self._months
    
    @months.setter
    def months(self, mnths: list[int]):
        self._months = mnths

    @property
    def days(self):
        return self._days
    
    @days.setter
    def days(self, d):
        self._days = d

    @property
    def hours(self):
        return self._hours
    
    @hours.setter
    def hours(self, hrs: list[int]):
        self._hours = hrs

    @property
    def seasons(self):
        return self._seasons
    
    @seasons.setter
    def seasons(self, season_list: list[dict[str,list[int]]], combine='first'):
        self._seasons = season_list
        self._seasons_combine = combine

    def add_season(self, name: str, months: list[int]):
        self._seasons.append(
            {'name':name,
             'months': sorted(months)}
        )

    @property
    def season_names(self):
        return [s['name'] for s in self.seasons]

    @property
    def dataframe(self):
        return self._dataframe

    def retrieve(self, N: int, method: _ENSEMBLE_TYPES='randomize', processes=1):

        if not self.years:
            years = list(self.years_available)
        else:
            years = [y for y in self.years if y in self.years_available]
            skipped_years = [y for y in self.years if y not in years]
            if skipped_years:
                print_text(f"skipping years {[print(y) for y in skipped_years]}")
        
        if not self.months:
            months = list(self.months_available)
        else:
            months = [m for m in self.months if m in self.months_available]
            skipped_months = [m for m in self.months if m not in months]
            if skipped_months:
                print_text(f"skipping months {[print(m) for m in skipped_months]}")

        if not self.days: # No days are set
            self.days = [d for d in range(1,32)]

        if not self.hours:
            self.hours = [0]

        datetimes = []
        for j in range(N):
            match method:
                case 'randomize':
                    year = random.choice(years)
                    month = random.choice(months)
                    if month==2:
                        day = random.choice(self.days[:28])
                    elif month in [4,6,9,11]:
                        day = random.choice(self.days[:-2])
                    else:
                        day = random.choice(self.days)
                    hour = random.choice(self.hours)
                case 'sequential':
                    year = years
            datetimes.append(datetime.fromisoformat(f"{year:04d}{month:02d}{day:02d} {hour:02d}00"))

        out = []
        pool = Pool(processes=processes)
        out = pool.map(self._build_met_list, datetimes)
        out = [x for sublist in out for x in sublist]

        met_data = pd.concat(out, ignore_index=True)

        self._apply_seasons_to_dataframe(met_data)

        return met_data


    def _apply_seasons_to_dataframe(self, df):        
        if len(self._seasons)>0:
            df['season'] = df.apply(lambda x: self._get_seasons(x), axis=1)
        return df
    
    def _get_seasons(self, x: pd.core.series.Series, method: Literal['combine','first']='combine'):
        xx = [s['name'] for s in self.seasons if x['month'] in s['months']]
        if xx:
            if method=='combine':
                return ' & '.join(xx)
            else:
                return xx[0]
        else:
            return 'other'
    

    def _build_met_list(self, dt: datetime) -> list[pd.DataFrame]:
        out = []
        this_met = self._get_met_as_df(dt)
        out.append(this_met)
        return out


    def _get_met_as_df(self, dt: datetime) -> pd.DataFrame:
        filename = f"{dt.year:04d}{dt.month:02d}.{self._filetype}"
        met = NetcdfMet(os.path.join(self._dirc,filename))
        met.extract(self.latitude, self.longitude, dt, convention="to")
        met_df = pd.DataFrame(columns=(
            'year',
            'month',
            'day',
            'hour',
            'altitude',
            'pressure',
            'temperature',
            'relative_humidity',
            'density',
            'wind_u',
            'wind_v',
            'wind_speed',
            'wind_direction'))
        met_df.loc[0,'year'] = dt.year
        met_df.loc[0,'month'] = dt.month
        met_df.loc[0,'day'] = dt.day
        met_df.loc[0,'hour'] = dt.hour
        met_df.loc[0,'altitude'] = met.altitude
        met_df.loc[0,'pressure'] = met.pressure
        met_df.loc[0,'temperature'] = met.temperature
        met_df.loc[0,'relative_humidity'] = met.relhum
        met_df.loc[0,'density'] = met.density
        met_df.loc[0,'wind_u'] = met.wind_U
        met_df.loc[0,'wind_v'] = met.wind_V
        met_df.loc[0,'wind_speed'] = met.wind_speed
        met_df.loc[0,'wind_direction'] = met.wind_direction
        met.close()
        return met_df


def _plot_windrose_subplots(data, *, direction, var, color=None, **kwargs):
    """wrapper function to create subplots per axis"""
    ax = plt.gca()
    ax = wr.WindroseAxes.from_ax(ax=ax)
    wr.plot_windrose(direction_or_df=data[direction], var=data[var], ax=ax, **kwargs)



def plot_windroses(data: MetRepo | pd.DataFrame, altitudes: list[float], 
                   N: int=1000, 
                   title: Optional[str]=None,
                   row: Optional[str]='altitude',
                   col: Optional[int]=None,
                   col_order: Optional[list[str]]=None,
                   include_all: bool=True,
                   method: _ENSEMBLE_TYPES='randomize', 
                   isblowto: bool=True,
                   processes=1):

    if isinstance(data, MetRepo):
        dataset = data.retrieve(N, method=method, processes=processes)
    elif isinstance(data, pd.DataFrame):
        dataset = data.copy()
    else:
        raise ValueError('plot_windroses data should be either MetRepo or pandas.DataFrame')
    
    df = pd.DataFrame(columns=('altitude','speed','direction'))
    for alt in altitudes:

        dataset[f'wind speed at {alt} m'] = dataset.apply(lambda x: np.interp(alt,x['altitude'],x['wind_speed']), axis=1)
        dataset[f'wind direction at {alt} m'] = dataset.apply(lambda x: np.interp(alt,x['altitude'],x['wind_direction']), axis=1)

        this_df = pd.DataFrame(columns=('altitude','speed','direction'))
        this_df['speed'] = dataset[f'wind speed at {alt} m']
        this_df['direction'] = dataset[f'wind direction at {alt} m']
        this_df['altitude'] = alt
        if col is not None:
            this_df[col] = dataset[col]

        if len(df)==0:
            df = this_df.copy()
        else:
            df = pd.concat([df, this_df], ignore_index=True)

    if include_all:
        df1 = df.copy()
        df1[col] = 'all'

        df = pd.concat([df,df1])

    max_spd = df.apply(lambda x: np.max(x['speed']), axis=1).max()
    spd_bins = [0.1] + np.linspace(1,max_spd,5).tolist()

    g = sns.FacetGrid(
        data=df,
        row=row,
        col=col,
        col_order=col_order,
        subplot_kws={"projection":"windrose"},
        sharex=False,
        sharey=False,
        despine=False,
        height=3.5,
    )

    g.map_dataframe(
        _plot_windrose_subplots,
        direction="direction",
        var="speed",
        normed=True,
        calm_limit=0.1,
        kind="bar",
        opening=1.0,
        bins=spd_bins,
        blowto=not isblowto,
        # edgecolor='k',
    )

    rows = df[row].unique()
    if col_order is not None:
        cols = col_order
    else:
        cols = df[col].unique()

    g.set_titles(template='')

    if col is not None:
        for ax, c in zip(g.axes[0], cols):
            c_title = f"{col}: {c}"
            ax.annotate(c_title, xy=(0.5,1), 
                        xytext=(0,5), 
                        xycoords='axes fraction', 
                        textcoords='offset points', 
                        ha='center', 
                        va='baseline',
                        size='large')

    for ax, r in zip(g.axes[:,0], rows):
        if row=='altitude':
            r_title = f"{row} = {r} m"
        else:
            r_title = f"{row}: {r}"
        ax.annotate(r_title, xy=(0,0.5), 
                    xytext=(-ax.yaxis.labelpad - 5,0), 
                    xycoords=ax.yaxis.label, 
                    textcoords='offset points', 
                    ha='right',
                    va='center',
                    size='large',
                    rotation=90)


    # if row is not None:
    #     if col=='altitude':
    #         g.set_titles(template="{col_name} m")
    #     elif row=='altitude':
    #         g.set_titles(template="{row_name} m")
    g.set_xticklabels([])
    max_wd_freq = 0 
    for ax in g.axes.flatten():
        table = ax._info["table"]
        wd_freq = np.sum(table, axis=0)
        max_wd_freq = np.maximum(max_wd_freq, np.amax(wd_freq))
    
    freq_ticks = np.linspace(0,np.ceil(max_wd_freq/5)*5,6, dtype=int)
    freq_ticks = freq_ticks[1:]
    for ax in g.axes.flatten():
        # title = ax.get_title()
        # ax.set_title('')
        # ax.set_title(title, loc='left')
        # dirc_label = ax.get_xticklabels()
        ax.set_rgrids(freq_ticks, freq_ticks)
        ax.set(rlabel_position=-90)
        ax.tick_params(axis='both', labelsize=8, pad=0)
        

    g.axes.flatten()[-1].set_legend(ncols=1, bbox_to_anchor=(1.1,1,0.1,0.5), title='Wind speed (m/s)')
    for l in g.axes.flatten()[-1].get_legend().texts:
        lt = l.get_text()
        lt = lt.replace('[','')
        lt = lt.replace(')','')
        lt = lt.replace(':',u'\u2014')
        l.set_text(lt)

    fig = plt.gcf()
    fig.subplots_adjust(right=0.8, wspace=0.1, hspace=0.1)

    return fig, g






from .epsg_defs import webmerc, wgs84
from .mapping import (add_north_arrow, add_opentopo_basemap, add_POI_marker,
                      add_scale_bar, ax_ticks, latlon_to_utm_epsg,
                      set_figure_size, set_min_axes, stamen_zoom_level)
from .webmap import BindColormap



from .interface import (load_inputs, save_inputs, set_grain_parameters,
                        set_met, set_met_parameters, set_model_parameters,
                        set_output_parameters, set_parameters,
                        set_physical_parameters, set_solver_parameters,
                        set_source_parameters, 
                        estimate_dispersal_distance, advected_settling_trajectories)



# import importlib.resources as pkg_resources
from importlib import resources
import pandas as pd

# gvpList = pkg_resources.open_text('data', "GVP_Volcano_List.csv")

def get_gvp_list():

    gvpList = resources.files(__package__).joinpath("GVP_Volcano_List.csv").open('r')

    gvp_list = pd.read_csv(gvpList)
    gvp_list['Search_Name'] = gvp_list['Volcano_Name']
    gvp_list['Primary_Name'] = ''
    gvp_list['Primary_indx'] = None
    gvp_list['Primary_Volcano'] = False
    gvp_list.loc[(~pd.isna(gvp_list['Latitude'])) & (~pd.isna(gvp_list['Longitude'])), 'Primary_Volcano'] = True
    for j, row in gvp_list.iterrows():
        if not row.Primary_Volcano:
            volc_num = row.Volcano_Number
            primary_indx = gvp_list.loc[(gvp_list.Volcano_Number==volc_num) & (gvp_list.Primary_Volcano==True)].index[0]
            primary_volc = gvp_list.loc[primary_indx,'Volcano_Name']
            gvp_list.loc[j,'Primary_Name'] = primary_volc
            gvp_list.loc[j,'Search_Name'] += f" ({primary_volc})"
            gvp_list.loc[j,'Primary_indx'] = primary_indx
        else:
            gvp_list.loc[j,'Primary_Name'] = row.Volcano_Name
            gvp_list.loc[j,'Primary_indx'] = j

    return gvp_list

def get_gvp_primary_list() -> pd.DataFrame:
    gvp_list = get_gvp_list()
    gvp_primary_list = gvp_list.dropna(subset=['Latitude','Longitude'])

    return gvp_primary_list


from collections import OrderedDict
from math import ceil

import numpy as np
from numba import float64
from numba.experimental import jitclass
from numba.typed import List
from numba.types import ListType

model_spec = OrderedDict()
# Scales and dimensionless parameters as arrays, one for each grain size
model_spec["SettlingScale"] = ListType(float64)  # Settling speed
model_spec["Velocity_ratio"] = ListType(float64)  # wind speed/settling speed
# model_spec["xyScale"] =ListType(t64[::1)]  # Horizontal scale
model_spec["xScale"] = ListType(float64)  # Easting scale
model_spec["yScale"] = ListType(float64)  # Northing scale
model_spec["Lx"] = ListType(float64)  # Dimensionless extent in x
model_spec["Ly"] = ListType(float64)  # Dimensionless extent in y
model_spec["cScale"] = ListType(float64)  # Concentration
model_spec["QScale"] = ListType(float64)  # Source term
model_spec["Peclet_number"] = float64 # Peclet number
model_spec["Diffusion_ratio"] = float64 # Kappa_h/Kappa_v
model_spec["sigma_hat"] = ListType(float64)  # source radius/xyScale
model_spec["sigma_hat_scale"] = ListType(float64)  # sigma_hat scaled to -pi->pi


@jitclass(model_spec)
class ModelParameters:

    # pylint: disable=too-many-instance-attributes
    # These attributes are all needed in ModelParameters

    def __init__(self):
        self.SettlingScale = List.empty_list(np.float64)
        self.Velocity_ratio = List.empty_list(np.float64)
        self.xScale = List.empty_list(np.float64)
        self.yScale = List.empty_list(np.float64)
        self.Lx = List.empty_list(np.float64)
        self.Ly = List.empty_list(np.float64)
        self.cScale = List.empty_list(np.float64)
        self.QScale = List.empty_list(np.float64)
        self.Peclet_number = np.nan
        self.Diffusion_ratio = np.nan
        self.sigma_hat = List.empty_list(np.float64)
        self.sigma_hat_scale = List.empty_list(np.float64)

        
    def from_params(
        self, params, xScale, yScale,
    ):
        solver_params = params.solver
        met_params = params.met
        source_params = params.source
        grain_params = params.grains
        physical_params = params.physical
        
        N = met_params.Ws_scale.shape[0]
        
        Peclet_number = (
            physical_params.Kappa_h / met_params.U_scale / source_params.PlumeHeight
        )
        Diffusion_ratio = physical_params.Kappa_h / physical_params.Kappa_v
        self.Peclet_number = Peclet_number
        self.Diffusion_ratio = Diffusion_ratio
        for j in range(N):
            SettlingScale = met_params.Ws_scale[j]
            sigma_hat = source_params.radius / np.maximum(xScale[j], yScale[j])
            self._add_to_list(
                SettlingScale,
                met_params.U_scale / SettlingScale,
                xScale[j],
                yScale[j],
                ceil(3 * source_params.radius / xScale[j]) * solver_params.domX,
                ceil(3 * source_params.radius / yScale[j]) * solver_params.domY,
                grain_params.proportion[j] * source_params.MER / source_params.radius**2 / SettlingScale,
                grain_params.proportion[j] * source_params.MER,
                sigma_hat,
                np.pi * sigma_hat / np.sqrt(solver_params.domX * solver_params.domY)
            )

    def from_values(
        self,
        SettlingScale,
        Velocity_ratio,
        xScale,
        yScale,
        Lx,
        Ly,
        cScale,
        QScale,
        Peclet_number,
        Diffusion_ratio,
        sigma_hat,
        sigma_hat_scale,
    ):
        self.Diffusion_ratio = np.float64(Diffusion_ratio)
        self.Peclet_number = np.float64(Peclet_number)
        self._add_to_list(
            SettlingScale,
            Velocity_ratio,
            xScale,
            yScale,
            Lx,
            Ly,
            cScale,
            QScale,
            sigma_hat,
            sigma_hat_scale,
        )

    def from_lists(self,
        SettlingScale,
        Velocity_ratio,
        xScale,
        yScale,
        Lx,
        Ly,
        cScale,
        QScale,
        Peclet_number,
        Diffusion_ratio,
        sigma_hat,
        sigma_hat_scale,
    ):
        """Initialize model parameters from lists"""

        self.Diffusion_ratio = np.float64(Diffusion_ratio)
        self.Peclet_number = np.float64(Peclet_number)

        for (W, V, xS, yS, lx, ly, cS, QS, sh, shS) in zip(SettlingScale,
            Velocity_ratio,
            xScale,
            yScale,
            Lx,
            Ly,
            cScale,
            QScale,
            sigma_hat,
            sigma_hat_scale
        ):
            self._add_to_list(W, V, xS, yS, lx, ly, cS, QS, sh, shS)

        return
        

    def _add_to_list(self,
        SettlingScale,
        Velocity_ratio,
        xScale,
        yScale,
        Lx,
        Ly,
        cScale,
        QScale,
        sigma_hat,
        sigma_hat_scale,
    ):
        self.SettlingScale.append(np.float64(SettlingScale))
        self.Velocity_ratio.append(np.float64(Velocity_ratio))
        self.xScale.append(np.float64(xScale))
        self.yScale.append(np.float64(yScale))
        self.Lx.append(np.float64(Lx))
        self.Ly.append(np.float64(Ly))
        self.cScale.append(np.float64(cScale))
        self.QScale.append(np.float64(QScale))
        self.sigma_hat.append(np.float64(sigma_hat))
        self.sigma_hat_scale.append(np.float64(sigma_hat_scale))

    def describe(self):
        print("Model parameters for AshDisperse")
        print("  Settling speed scale = ", self.SettlingScale)
        print("  Velocity ratio = ", self.Velocity_ratio)
        print("  concentration scale = ", self.cScale)
        # print("  x and y scale = ", self.xyScale)
        print("  x scale = ", self.xScale)
        print("  y scale = ", self.yScale)
        print("  Lx = ", self.Lx)
        print("  Ly = ", self.Ly)
        print("  source flux scale = ", self.QScale)
        print("  Peclet number = ", self.Peclet_number)
        print("  Diffusion ratio = ", self.Diffusion_ratio)
        print("********************")


# pylint: disable=E1101
ModelParameters_type = ModelParameters.class_type.instance_type



from collections import OrderedDict

import numpy as np
from numba import boolean, float64, int64
from numba.experimental import jitclass

solver_spec = OrderedDict()
solver_spec["domX"] = float64  # dimensionless domain length in x
solver_spec["domY"] = float64  # dimensionless domain length in y
solver_spec["minN_log2"] = int64  # log2 minimum number of Chebyshev points
solver_spec["maxN_log2"] = int64  # log2 maximum number of Chebyshev points
solver_spec["Nx_log2"] = int64  # log2 of number of points in x
solver_spec["Ny_log2"] = int64  # log2 of number of points in y
solver_spec["epsilon"] = float64  # tolerance for converged spectral series
solver_spec["fft_tol"] = float64  # tolerance for fft terms
solver_spec["meps"] = float64  # Machine epsilon


@jitclass(solver_spec)
class SolverParameters:
    def __init__(
        self,
        domX=1.5,
        domY=1.5,
        minN_log2=4,
        maxN_log2=8,
        Nx_log2=8,
        Ny_log2=8,
        epsilon=1e-8,
        fft_tol=1e-10,
    ):
        self.meps = np.finfo(np.float64).eps

        self.domX = np.float64(domX)  # Dimensionless domain size in x
        self.domY = np.float64(domY)  # Dimensionless domain size in y

        self.minN_log2 = np.int64(minN_log2)  # Minimum z-resolution (log2)
        self.maxN_log2 = np.int64(maxN_log2)  # Maximum z-resolution (log2)

        self.Nx_log2 = np.int64(Nx_log2)  # x-resolution (log2)
        self.Ny_log2 = np.int64(Ny_log2)  # y-resolution (log2)

        self.epsilon = np.float64(epsilon)
        self.fft_tol = np.float64(fft_tol)

    def validate(self):
        if self.domX < 0:
            raise ValueError("In SolverParameters, must have domX>0")
        if self.domY < 0:
            raise ValueError("In SolverParameters, must have domY>0")
        if self.minN_log2 < 0:
            raise ValueError("In SolverParameters, must have minN_log2>0")
        if self.maxN_log2 < 0:
            raise ValueError("In SolverParameters, must have maxN_log2>0")
        if self.minN_log2 > self.maxN_log2:
            raise ValueError("In SolverParameters, must have minN_log2 < maxN_log2")
        if self.Nx_log2 < 0:
            raise ValueError("In SolverParameters, must have Nx_log2>0")
        if self.Ny_log2 < 0:
            raise ValueError("In SolverParameters, must have Ny_log2>0")
        if self.epsilon < 0:
            raise ValueError("In SolverParameters, must have epsilon>0")
        if self.epsilon < 10 * self.meps:
            raise ValueError(
                f"In SolverParameters, must have epsilon >= 10*machine epsilon = {10 * self.meps}"
            )
        if self.fft_tol < 0:
            raise ValueError("In SolverParameters, must have fft_tol>0")
        if self.fft_tol < 10 * self.meps:
            raise ValueError(
                f"In SolverParameters, must have fft_tol >= 10*machine epsilon = {10 * self.meps}"
            )
        return 1        

    @property
    def Nx(self):
        return 2**self.Nx_log2
    
    @property
    def Ny(self):
        return 2**self.Ny_log2

    @property
    def minN(self):
        return 2**self.minN_log2
    
    @property
    def maxN(self):
        return 2**self.maxN_log2
    
    @property
    def chebIts(self):
        return self.maxN_log2 - self.minN_log2 + 1

    def describe(self):
        print("Solver parameters for AshDisperse")
        print("  Dimensionless domain size in x, domX = ", self.domX)
        print("  Dimensionless domain size in y, domY = ", self.domY)
        print(
            "  Minimum resolution in z, minN = ",
            self.minN,
            " (minN_log2 = ",
            self.minN_log2,
            ")",
        )
        print(
            "  Maximum resolution in z, maxN = ",
            self.maxN,
            " (maxN_log2 = ",
            self.maxN_log2,
            ")",
        )
        print("  Number of Chebyshev iterates = ", self.chebIts)
        print("  Tolerance for Chebyshev series, epsilon = ", self.epsilon)
        print("  Tolerance for FFT terms, fft_tol = ", self.fft_tol)
        print("  Resolution in x, Nx = ", self.Nx, " (Nx_log2 = ", self.Nx_log2, ")")
        print("  Resolution in y, Ny = ", self.Ny, " (Ny_log2 = ", self.Ny_log2, ")")
        print("********************")


# pylint: disable=E1101
SolverParameters_type = SolverParameters.class_type.instance_type

def _solver_dict(p):
    return {
        'domX': float(p.domX),
        'domY': float(p.domY),
        'minN_log2': int(p.minN_log2),
        'maxN_log2': int(p.maxN_log2),
        'Nx_log2': int(p.Nx_log2),
        'Ny_log2': int(p.Ny_log2),
        'epsilon': float(p.epsilon),
        'fft_tol': float(p.fft_tol),
    }

def _solver_params_equal(p1: SolverParameters, p2: SolverParameters) -> bool:
    test = (
        (p1.domX == p2.domX) and
        (p1.domY == p2.domY) and
        (p1.minN_log2 == p2.minN_log2) and
        (p1.maxN_log2 == p2.maxN_log2) and
        (p1.Nx_log2 == p2.Nx_log2) and
        (p1.Ny_log2 == p2.Ny_log2) and
        (p1.epsilon == p2.epsilon) and
        (p1.fft_tol == p2.fft_tol)
    )
    return test


import subprocess

from pkg_resources import resource_filename


def run_notebook(file):
    subprocess.Popen(["jupyter notebook " + file], shell=True)


def launch_jupyter_example():
    file = resource_filename(__name__, "/ashdisperse.ipynb")
    print(f"Running {file}")
    run_notebook(file)



import datetime
import os
from collections import OrderedDict
from typing import Optional

import numpy as np
import toml
from numba import optional
from numba.experimental import jitclass
from numba.typed import List

from ..queryreport import is_valid_lat, is_valid_lon
from ..utilities import latlon_point_to_utm_code
from .emission_params import (EmissionParameters, EmissionParameters_type,
                              Suzuki_k_from_peak, _emission_dict)
from .grain_params import GrainParameters, GrainParameters_type, _grains_dict
from .met_params import MetParameters, MetParameters_type
from .model_params import ModelParameters, ModelParameters_type
from .output_params import (OutputParameters, OutputParameters_type,
                            _output_dict)
from .physical_params import (PhysicalParameters, PhysicalParameters_type,
                              _physical_dict)
from .solver_params import (SolverParameters, SolverParameters_type,
                            _solver_dict)
from .source_params import (SourceParameters, SourceParameters_type,
                            _source_dict)

param_spec = OrderedDict()
param_spec["solver"] = optional(SolverParameters_type)
param_spec["grains"] = optional(GrainParameters_type)
param_spec["emission"] = optional(EmissionParameters_type)
param_spec["source"] = optional(SourceParameters_type)
param_spec["physical"] = optional(PhysicalParameters_type)
param_spec["met"] = optional(MetParameters_type)
param_spec["model"] = optional(ModelParameters_type)
param_spec["output"] = optional(OutputParameters_type)


@jitclass(param_spec)
class Parameters:
    def __init__(self):
        self.source = None
        self.grains = None
        self.emission = None
        self.solver = None
        self.physical = None
        self.met = None
        self.model = None
        self.output = None

    def update(self, with_utm: bool=True):
        if with_utm:
            self.source.utm_from_latlon()
        _ = self.source.validate()
        _ = self.grains.validate()
        _ = self.emission.validate()
        _ = self.solver.validate()
        _ = self.physical.validate()
        _ = self.met.validate()
        _ = self.output.validate()

        self.output.ChebMats(self.solver.maxN, self.source.PlumeHeight)

    def describe(self):
        print("AshDisperse parameters")
        self.source.describe()
        self.grains.describe()
        self.emission.describe()
        self.solver.describe()
        self.physical.describe()
        if self.model is not None:
            self.model.describe()
        # self.met.describe()
        self.output.describe()


# pylint: disable=E1101
Parameters_type = Parameters.class_type.instance_type


def copy_parameters(A):
    new_A = Parameters()
    new_A.source = SourceParameters(
        A.source.latitude,
        A.source.longitude,
        A.source.utmcode,
        A.source.radius,
        A.source.PlumeHeight,
        A.source.MER,
        A.source.duration,
        name=A.source.name,
    )
    new_A.grains = GrainParameters()
    for j in range(A.grains.bins):
        new_A.grains.add_grain(
            A.grains.diameter[j], A.grains.density[j], A.grains.proportion[j]
        )
    new_A.emission = EmissionParameters()
    for j in range(A.emission.len()):
        new_A.emission.add_profile(
            A.emission.lower[j], A.emission.upper[j], A.emission.profile[j], A.emission.Suzuki_k[j]
        )
    new_A.solver = SolverParameters(
        domX=A.solver.domX,
        domY=A.solver.domY,
        minN_log2=A.solver.minN_log2,
        maxN_log2=A.solver.maxN_log2,
        epsilon=A.solver.epsilon,
        fft_tol=A.solver.fft_tol,
        Nx_log2=A.solver.Nx_log2,
        Ny_log2=A.solver.Ny_log2,
    )
    new_A.physical = PhysicalParameters(
        A.physical.Kappa_h, A.physical.Kappa_v, A.physical.g, A.physical.mu
    )
    if A.met is not None:
        new_A.met = MetParameters(A.met.U_scale, A.met.Ws_scale)
    if A.model is not None:
        new_A.model = ModelParameters()
        new_A.model.from_lists(
            A.model.SettlingScale,
            A.model.Velocity_ratio,
            A.model.xScale,
            A.model.yScale,
            A.model.Lx,
            A.model.Ly,
            A.model.cScale,
            A.model.QScale,
            A.model.Peclet_number,
            A.model.Diffusion_ratio,
            A.model.sigma_hat,
            A.model.sigma_hat_scale,
        )
    new_A.output = OutputParameters(A.output.start, A.output.stop, A.output.step)
    new_A.output.set_altitudes()
    new_A.output.ChebMats(A.solver.maxN, A.source.PlumeHeight)
    return new_A


def update_parameters(
    A,
    name=None,
    domX=None,
    domY=None,
    minN_log2=None,
    maxN_log2=None,
    epsilon=None,
    fft_tol=None,
    Nx_log2=None,
    Ny_log2=None,
    grains=None,
    emissions: Optional[ list[dict[str,int | float]] | str ] = None,
    latitude=None,
    longitude=None,
    radius=None,
    PlumeHeight=None,
    MER=None,
    duration=None,
    Kappa_h=None,
    Kappa_v=None,
    g=None,
    mu=None,
    start=None,
    stop=None,
    step=None,
):

    # solver
    if domX is not None:
        A.solver.domX = domX
    if domY is not None:
        A.solver.domY = domY
    if minN_log2 is not None:
        A.solver.minN_log2 = minN_log2
    if maxN_log2 is not None:
        A.solver.maxN_log2 = maxN_log2
    if epsilon is not None:
        A.solver.epsilon = epsilon
    if fft_tol is not None:
        A.solver.fft_tol = fft_tol
    if Nx_log2 is not None:
        A.solver.Nx_log2 = Nx_log2
    if Ny_log2 is not None:
        A.solver.Ny_log2 = Ny_log2

    # grains
    if grains is not None:
        if not isinstance(grains, list):
            raise ValueError(
                "in update_parameters, grains must be a list of dicts; "
                + "received {}".format(grains)
            )
        if not all(
            "class" in g and "diameter" in g and "density" in g and "proportion" in g
            for g in grains
        ):
            raise ValueError(
                "in update parameters, grains must be a list of dicts with "
                + "dict containing the keys 'class', 'diameter', 'density',"
                + " 'proportion'; received {}".format(grains)
            )
        grains = sorted(grains, key=lambda p: p["class"], reverse=False)
        diameter = [g['diameter'] for g in grains]
        density = [g['density'] for g in grains]
        proportion = [g['proportion'] for g in grains]
        A.grains.from_lists(diameter, density, proportion)
        A.grains.validate()

    if emissions is not None:
        if isinstance(emissions, list):
            if not all(
                "class" in g and "lower" in g and "upper" in g and ("Suzuki_k" in g or "Suzuki_peak" in g)
                for g in emissions
            ):
                raise ValueError(
                    "in update parameters, emissions must be a list of dicts with "
                    + "dict containing the keys 'class', 'lower', 'upper',"
                    + f" and either 'Suzuki_k' or 'Suzuki_peak'; received {emissions}"
                )
            emissions = sorted(emissions, key=lambda p: p["class"], reverse=False)
            profile_list = []
            lower_list = []
            upper_list = []
            Suzuki_k_list = []
            for this_emission in emissions:
                profile = 0
                lower = this_emission["lower"]
                upper = this_emission["upper"]
                if "Suzuki_peak" in this_emission:
                    Suzuki_k = Suzuki_k_from_peak(this_emission["Suzuki_peak"],lower,upper)
                else:
                    Suzuki_k = this_emission["Suzuki_k"]
                profile_list.append(profile)
                lower_list.append(lower)
                upper_list.append(upper)
                Suzuki_k_list.append(Suzuki_k)
            
            A.emission.from_lists(lower_list, upper_list, profile_list, Suzuki_k_list)
        elif isinstance(emissions, str):
            if emissions == 'single':
                profile_list = [A.emission.profile[0] for _ in range(A.grains.bins)]
                lower_list = [A.emission.lower[0] for _ in range(A.grains.bins)]
                upper_list = [A.emission.upper[0] for _ in range(A.grains.bins)]
                Suzuki_k_list = [A.emission.Suzuki_k[0] for _ in range(A.grains.bins)]

                A.emission.from_lists(lower_list, upper_list, profile_list, Suzuki_k_list)
            else:
                raise ValueError(f"in update_parameters, emissions must be a list of dicts or the string 'single'; "
                    + f"received {emissions}")

        else:
            raise ValueError(
                f"in update_parameters, emissions must be a list of dicts or the string 'single'; "
                + f"received {emissions}"
            )
    
    if not (grains is not None) == (emissions is not None):
        raise ValueError("Need to update both grains and emissions")

    # source
    if name is not None:
        A.source.name = name
    if latitude is not None:
        if is_valid_lat(latitude):
            A.source.latitude = np.float64(latitude)
        else:
            raise ValueError(
                "In update_parameters, latitude must be a valid latitude in "
                + " decimal degrees"
            )
    if longitude is not None:
        if is_valid_lon(longitude):
            A.source.longitude = np.float64(longitude)
        else:
            raise ValueError(
                "In update_parameters, latitude must be a valid longitude in "
                + " decimal degrees"
            )
    if (latitude is not None) or (longitude is not None):
        A.source.utmcode = latlon_point_to_utm_code(
            A.source.latitude, A.source.longitude
        )

    if radius is not None:
        if radius < 0:
            raise ValueError("In update_parameters, radius must be positive")
        A.source.radius = np.float64(radius)

    if PlumeHeight is not None:
        if PlumeHeight < 0:
            raise ValueError("In update_parameters, PlumeHeight must be positive")
        A.source.PlumeHeight = np.float64(PlumeHeight)

    if MER is not None:
        if MER < 0:
            raise ValueError("In update_parameters, MER must be positive")
        A.source.MER = np.float64(MER)

    if duration is not None:
        if duration < 0:
            raise ValueError("In update_parameters, duration must be positive")
        A.source.duration = np.float64(duration)

    # physical
    if Kappa_h is not None:
        if Kappa_h < 0:
            raise ValueError("In PhysicalParameters, Kappa_h must be positive")
        A.physical.Kappa_h = np.float64(Kappa_h)

    if Kappa_v is not None:
        if Kappa_v < 0:
            raise ValueError("In PhysicalParameters, Kappa_v must be positive")
        A.physical.Kappa_v = np.float64(Kappa_v)

    if g is not None:
        if g <= 0:
            raise ValueError("In PhysicalParameters, g must be positive")
        A.physical.g = np.float64(g)

    if mu is not None:
        if mu <= 0:
            raise ValueError("In PhysicalParameters, mu must be positive")
        A.physical.mu = np.float64(mu)

    # wind_speed = met.max_wind_speed(A.source.PlumeHeight)
    # ws = met.settling_speed_value(A, 0.0)
    # A.met = MetParameters(wind_speed, ws)

    # A.model.from_params(A.solver, A.met, A.source, A.grains, A.physical)

    if start is not None:
        A.output.start = np.float64(start)
    if stop is not None:
        A.output.stop = np.float64(stop)
    if step is not None:
        A.output.step = np.float64(step)

    A.output.set_altitudes()
    A.output.ChebMats(A.solver.maxN, A.source.PlumeHeight)

    return A


def save_parameters(params, file="parameters.toml"):
    if os.path.exists(file):
        print(
            "WARNING: {outname} already exists and will be replaced".format(
                outname=file
            )
        )

    pdict = {
        "source": _source_dict(params.source),
        "grains": _grains_dict(params.grains),
        "emission": _emission_dict(params.emission),
        "solver": _solver_dict(params.solver),
        "physical": _physical_dict(params.physical),
        "output": _output_dict(params.output),
    }
    
    with open(file, "w") as f:
        toml.dump(pdict,f)


def load_parameters(file):
    if not os.path.exists(file):
        raise IOError("AshDisperse parameters file {} not found".format(file))

    try:
        with open(file, 'r') as f:
            paramset = toml.load(f)
    except:
        raise RuntimeError(f'Unable to read file {file}')

    params = parameters_from_dict(paramset)

    # windSpeed = met.max_wind_speed(params.source.PlumeHeight)
    # ws = met.settling_speed(params, np.atleast_1d(0.0))[0]
    # params.met = MetParameters(windSpeed, ws)

    # params.model = ModelParameters()
    # params.model.from_params(params.solver,
    #                          params.met,
    #                          params.source,
    #                          params.grains,
    #                          params.physical)

    return params

def parameters_from_dict(paramset: dict) -> Parameters_type:

    params = Parameters()

    params.source = SourceParameters(
        paramset["source"]["latitude"],
        paramset["source"]["longitude"],
        paramset["source"]["utmcode"],
        radius=paramset["source"]["radius"],
        PlumeHeight=paramset["source"]["PlumeHeight"],
        MER=paramset["source"]["MER"],
        duration=paramset["source"]["duration"],
        name=paramset["source"]["name"],
    )

    params.grains = GrainParameters()
    diameters = List()
    [diameters.append(d) for d in paramset["grains"]["diameter"]]
    densities = List()
    [densities.append(p) for p in paramset["grains"]["density"]]
    props = List()
    [props.append(p) for p in paramset["grains"]["proportion"]]
    params.grains.from_lists(diameters, densities, props)
    
    params.emission = EmissionParameters()
    lower = List()
    [lower.append(d) for d in paramset["emission"]["lower"]]
    upper = List()
    [upper.append(p) for p in paramset["emission"]["upper"]]
    profile = List()
    # Backward compatibility -- depreciate in future...
    if 'profile' in paramset["emission"].keys():
        [profile.append(p) for p in paramset["emission"]["profile"]]
    else:
        [profile.append(0) for p in paramset["emission"]["upper"]]
    Suzuki_k = List()
    [Suzuki_k.append(p) for p in paramset["emission"]["Suzuki_k"]]
    params.emission.from_lists(lower, upper, profile, Suzuki_k)

    params.solver = SolverParameters(
        domX=paramset["solver"]["domX"],
        domY=paramset["solver"]["domY"],
        minN_log2=paramset["solver"]["minN_log2"],
        maxN_log2=paramset["solver"]["maxN_log2"],
        epsilon=paramset["solver"]["epsilon"],
        fft_tol=paramset["solver"]["fft_tol"],
        Nx_log2=paramset["solver"]["Nx_log2"],
        Ny_log2=paramset["solver"]["Ny_log2"],
    )

    
    params.physical = PhysicalParameters(
        Kappa_h=paramset["physical"]["Kappa_h"],
        Kappa_v=paramset["physical"]["Kappa_v"],
        g=paramset["physical"]["g"],
        mu=paramset["physical"]["mu"]
    )

    params.output = OutputParameters(
        start=paramset["output"]["start"],
        stop=paramset["output"]["stop"],
        step=paramset["output"]["step"],
    )
    params.output.set_altitudes()
    params.output.ChebMats(params.solver.maxN, params.source.PlumeHeight)

    return params




from numba import njit

webmerc = {'init': 'epsg:3857'}
wgs84 = {'init': 'epsg:4326'}

@njit
def latlon_point_to_utm_code(lat, lon):
    offset = int(round((183+lon)/6.0))
    return 32600+offset if (lat > 0) else 32700+offset



import cdsapi
from datetime import datetime
from math import ceil, floor
import metpy.units as metunits
import metpy.calc as metcalc
import numpy as np
import numpy.typing as npt
import os
from typing import Optional, Literal
import xarray as xr

from ..queryreport import print_text

class NetcdfMet:
    def __init__(self, file: str):
        self._file = file

        self._open = False

        self.altitude = np.empty((0), dtype=np.float64)
        self.wind_U = np.empty((0), dtype=np.float64)
        self.wind_V = np.empty((0), dtype=np.float64)
        self.wind_speed = np.empty((0), dtype=np.float64)
        self.wind_direction = np.empty((0), dtype=np.float64)
        self.temperature = np.empty((0), dtype=np.float64)
        self.pressure = np.empty((0), dtype=np.float64)
        self.density = np.empty((0), dtype=np.float64)

        self._time_coord = None


    def close(self):
        if self._open:
            self._data.close()
            self._open = False

    @property
    def file(self) -> str:
        return self._file
    
    @property
    def data(self) -> xr.Dataset:
        if not self._open:
            dat = xr.load_dataset(self._file, engine='netcdf4')
            dat = dat.metpy.parse_cf() # read metadata
            self._data = dat
            self._open = True
        return self._data

    def _get_time_coord(self):
        if 'valid_time' in self.data.coords:
            self._time_coord = 'valid_time'
        elif 'time' in self.data.coords:
            self._time_coord = 'time'
        else:
            self._time_coord = None
            Warning(f'time coordinate not recognized in {self._file}')
            

    @property
    def time_coord(self) -> str | None:
        if self._time_coord is None:
            self._get_time_coord()
        return self._time_coord

    @property
    def latitude(self) -> npt.NDArray:
        return self.data.latitude.values
    
    @property
    def longitude(self) -> npt.NDArray:
        return self.data.longitude.values
    
    @property
    def time(self) -> npt.NDArray:
        return self.data[self.time_coord].values
        
    @property
    def extent(self) -> list[float]:
        return [self.longitude.min(), self.longitude.max(), self.latitude.min(), self.latitude.max()]
    
    def _in_extent(self, lat: float, lon: float) -> bool:
        lon_min, lon_max, lat_min, lat_max = self.extent
        return (lon_min <= lon <= lon_max) and (lat_min <= lat <= lat_max)
    
    def _in_timespan(self, datetime: datetime) -> bool:
        return np.datetime64(datetime) in self.time

    def extract(self, lat: float, lon: float, datetime: datetime, convention: Literal["to", "from"]="to"):

        if not self._in_extent(lat, lon):
            print_text(f"Location {lat}, {lon} not in extent of {self.file}")
            return
        
        if not self._in_timespan(datetime):
            print_text(f"Time {datetime} not in timespan of {self.file}")
            return
                
        data0 = self.data.sel(valid_time=datetime).interp(latitude=lat, longitude=lon, method='cubic').metpy.quantify()

        geopot = data0['z']
        Z = metcalc.geopotential_to_height(geopot)
        Z = Z.metpy.convert_units("m")

        self.altitude = np.float64(Z.values)

        U = data0['u'].metpy.convert_units("m/s")
        V = data0['v'].metpy.convert_units("m/s")
        T = data0['t'].metpy.convert_units("K")
        RH = data0['r'].metpy.convert_units("%")
        P = data0['pressure_level']
        spd = metcalc.wind_speed(U,V)
        dir = metcalc.wind_direction(U,V, convention=convention)

        self.wind_U = np.float64(U.values)
        self.wind_V = np.float64(V.values)
        self.wind_speed = np.float64(spd.values)
        self.wind_direction = np.float64(dir.values)
        self.temperature = np.float64(T.values)
        self.relhum = np.float64(RH.values)
        self.pressure = np.float64(P.metpy.convert_units("Pa").values)

        mixing_ratio = metcalc.mixing_ratio_from_relative_humidity(P,T,RH)
        rho = metcalc.density(P,T,mixing_ratio)
        rho = rho.metpy.convert_units("kg/m^3")
        self.density = rho.values

    def download_era5(self, *, 
                 lat: Optional[float]=None, 
                 lon: Optional[float]=None, 
                 datetime: Optional[datetime]=None,
                 extent: Optional[list[float]]=None,
                 year: Optional[list[int]]=None,
                 month: Optional[list[int]]=None,
                 day: Optional[list[int]]=None,
                 hour: Optional[list[int]]=None):
        if os.path.isfile(self.file):
            print_text(f"Met file {self.file} exists and will be overwritten")

        if (lat is None) and (lon is None):
            if extent is None:
                    raise ValueError('set either lat and lon, or extent')
        else:
            if (lat is None) or (lon is None):
                raise ValueError('set both lat and lon')
            else:
                extent = [0.25*floor(lon/0.25), 
                          0.25*ceil(lon/0.25), 
                          0.25*floor(lat/0.25),
                          0.25*ceil(lat/0.25)]
        
        if datetime is not None:
            year = [datetime.year]
            month = [datetime.month]
            day = [datetime.day]
            hour = [datetime.hour]
        else:
            if (year is None) or (month is None) or (day is None) or (hour is None):
                raise ValueError('set either datetime or all of year, month, day and hour')

        cds_dataset = "reanalysis-era5-pressure-levels"
        cds_request = {
            "product_type": "reanalysis",
            "format": "netcdf",
            "variable": [
                "geopotential",
                "temperature",
                "relative_humidity",
                "u_component_of_wind",
                "v_component_of_wind",
            ],
            "pressure_level": [
                "1","2","3","5","7","10","20","30","50","70",
                "100","125","150","175","200","225","250","300",
                "350","400","450","500","550","600","650","700","750",
                "775","800","825","850","875","900","925","950","975","1000",
            ],
            "year": [str(yr) for yr in year],
            "month": ["{:02d}".format(mnth) for mnth in month],
            "day": ["{:02d}".format(d) for d in day],
            "time": ["{:02d}:00".format(hr) for hr in hour],
            "area": [
                extent[3], # max lat
                extent[0], # min lon
                extent[2], # min lat
                extent[1], # max lon
            ],
            "data_format": "netcdf",
            "download_format": "unarchived",
        }

        cds = cdsapi.Client()
        cds.retrieve(cds_dataset, cds_request).download(self.file)


from .config import (get_chunk_size, get_max_threads, get_num_threads,
                     get_threading_layer, set_chunk_size, set_default_threads,
                     set_num_threads, set_threading_layer)

set_threading_layer(thread_layer='tbb')
